CD-Title: LDM Related Functions (LFN)
CD-end

CD-Intro:
C These are sets of routines which provide some machine independent functions
C for inclusion in programs for processing Laue diffraction data. They are
C used in association with the Laue Data Module (LDM) routines.
CD-end

CD-Section_order:
C 1 3 4 2 6 5 7 8
CD-end

CD-Section: Crystal Orientation Refinement
C These routines are for refining the crystal orientation parameters and
C cell/distortion parameters as required.
CD-end

C LAST UPDATE: 
C+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
C
C**************************
C**     LFN_RFN          **
C**************************
C
C
CD-Routine: Refine Orientation Parameters - LFN_RFN
C This routine is used to refine the setting, cell and distortion parameters
C as desired. A refinement list must be set up using a call to LFN_CLROBS
C and calls to LFN_ADDOBS before the routine LFN_RFN is called.
CD-end
CD-Fortran:
      SUBROUTINE LFN_RFN (IPACK, IPLATE, IFLAG, JFLAG, KFLAG, 
     +                    X_CEN_F, Y_CEN_F, IRFTYP, TUNE, 
     +                    RMS, SOVEMX, SOVEAV, IERR, MESSAGE)
CD-end
C
      INCLUDE 'implicit_none.finc'
C
C Purpose: Refine setting, cell and distortion parameters
C
C          Note: Refinement spots list must be set up via a call to
C                LFN_CLROBS and calls to LFN_ADDOBS before calling
C                this routine
C
C Author:  John W. Campbell, September 1993
C
C Arguments:
C
      INTEGER IPACK
      INTEGER IPLATE
      INTEGER IFLAG(7)
      INTEGER JFLAG(6)
      INTEGER KFLAG(6)
      REAL X_CEN_F
      REAL Y_CEN_F
      INTEGER IRFTYP
      REAL TUNE(3)
      REAL RMS, SOVEMX, SOVEAV
      INTEGER IERR
      CHARACTER*(*) MESSAGE
CD-Parameters: 
C  IPACK       i (R)   Pack number
C  IPLATE      i (R)   Plate number
C  IFLAG()     i (R)   Parameter refinement flags 1=refine, 0=do not refine     
C                      IFLAG(1)  PHIX
C                      IFLAG(2)  PHIY
C                      IFLAG(3)  PHIZ
C                      IFLAG(4)  CTOF
C                      IFLAG(5)  X_C
C                      IFLAG(6)  Y_C
C                      IFLAG(7)  W_C
C  JFLAG()     i (R)   Cell parameter refinement flags for compressed
C                      cell (i.e. minimum cell parameters required for
C                      the crystal system (e.g. triclinic 6, 
C                      orthorhombic 3, cubic 1) 1=refine, 
C                      0=do not refine
C  KFLAG()     i (R)   Distortion parameter refinement flags
C
C                      For distortion type 1: KFLAG(1) Y_SCALE
C                                             KFLAG(2) TWIST
C                                             KFLAG(3) TILT
C                                             KFLAG(4) BULGE
C
C                      For distortion type 2: KFLAG(1) Y_SCALE
C                                             KFLAG(2) TWIST
C                                             KFLAG(3) TILT
C                                             KFLAG(4) ROFF
C                                             KFLAG(5) TOFF
C
C                      For distortion type 3: KFLAG(1) Y_SCALE
C                                             KFLAG(2) TWIST
C                                             KFLAG(3) TILT
C                                             KFLAG(4) SPDXY
C                                             KFLAG(5) SPDX
C                                             KFLAG(6) SPDY
C
C                      For distortion type 4: KFLAG(1) Y_SCALE
C                                             KFLAG(2) TWIST
C                                             KFLAG(3) TILT
C X_CEN_F       r (R)  Un-corrected x-centre position in rasters
C Y_CEN_F       r (R)  Un-corrected y-centre position in rasters
C IRFTYP        i (R)  Refinement type =1 LSR, =2 Powell
C TUNE(3)       r (R)  Refinement tuning parameters
C                      If IRFTYP = 1  TUNE(1) = shift limit e.g.0.001
C                                     TUNE(2) = step parameter e.g. 1.0E-10    
C                                     TUNE(3) = damping factor e.g. 0.5
C                       If IRFTYP = 2 TUNE(1) = accuracy e.g. 0.000001
C RMS           r (W)  rms deviation (mm) after refinement
C SOVEMX        r (W)  Max shift/error
C SOVEAV        r (W)  Avg shift/error
C IERR          i (W)  Error flag =0,  OK
C                                 =-1, Spatial distortion arrays cannot
C                                      handle all coefficients
C                                 =1,  Refinement not completed
C                                 =2,  Other refinement error
C                                 =3,  Spots found outside Chebyshev limits     
C MESSAGE       c (W)  Message string from refinement (up to 80 characters)
CD-end
C
C====== Parameter statements
C 
      INCLUDE 'refine_limits.finc'
      INTEGER LW1
      PARAMETER (LW1=5*MAX_N+2*MAX_N*MAX_N+2*MAX_M*MAX_N+2*MAX_M)
C
C LW1      Length of longest work array needed
C
C====== GLOBALS:
C
      INCLUDE 'refobs_list.finc'
      INCLUDE 'refine_data.finc'
C
C====== FUNCTION TYPES:
C
      EXTERNAL LFN_CALFUN
C
C====== LOCALS:
C
      INTEGER I
      INTEGER N
      INTEGER M
      INTEGER IFL
      INTEGER NCP(7)
      INTEGER LFLAG
      INTEGER ND
      INTEGER LERR
      INTEGER KERR
      INTEGER ISYST
      INTEGER DISTOR_TYPE
      INTEGER NSPDX
      INTEGER NSPDY
      REAL SPD_XY
      REAL SPDX_MIN
      REAL SPDX_MAX
      REAL SPDY_MIN
      REAL SPDY_MAX
      REAL SPDX(MAX_SPD)
      REAL SPDY(MAX_SPD)
      REAL REF_PAR(7)
      REAL RFCL_PAR(6)
      REAL DSTOR_PAR(2*MAX_SPD+4)
      REAL VALUE
      REAL CELL(6)
      REAL RCELL(6)
      REAL CELL_PARAMS(6)
      REAL SOVSUM, SOVN, R1
      REAL RTOD
      DOUBLE PRECISION X_VEC(MAX_N), E_VEC(MAX_N)
      DOUBLE PRECISION FMIN
      DOUBLE PRECISION FVECC(MAX_M)
      DOUBLE PRECISION W(LW1)
      DOUBLE PRECISION POWELL_ACC
      DOUBLE PRECISION LSR_SHFLM 
      DOUBLE PRECISION LSR_DSTEP
      DOUBLE PRECISION LSR_DAMP
      CHARACTER*2 VALSTR
      CHARACTER*2 ERRSTR
C
C     INTEGER I                      !Temp/loop variable
C     INTEGER N                      !No. of parameters to refine
C     INTEGER M                      !No. of obseravions in refinement
C     INTEGER IFL                    !Return flag from refinement routines
C     INTEGER NCP(7)                 !No. of cell params for each crystal system
C     INTEGER KFLAG                  !Parameter status flag
C     INTEGER ND                     !No. of decimal places flag
C     INTEGER LERR                   !Error flag
C     INTEGER KERR                   !Error flag
C     INTEGER ISYST                  !Crystal system
C     INTEGER DISTOR_TYPE            !Distortion type flag
C     INTEGER NSPDX                  !Spatial distortion x polynomial order
C     INTEGER NSPDY                  !Spatial distortion y polynomial order
C     REAL SPD_XY                    !Spatial distortion xy cross term
C     REAL SPDX_MIN                  !Spatial distortion chebyshev x lower bound
C     REAL SPDX_MAX                  !Spatial distortion chebyshev x upper bound
C     REAL SPDY_MIN                  !Spatial distortion chebyshev y lower bound
C     REAL SPDX_MAX                  !Spatial distortion chebyshev y upper bound
C     REAL SPDX(MAX_SPD)             !Spatial distortion x polynomial coeffts
C     REAL SPDY(MAX_SPD)             !Spatial distortion y polynomial coeffts
C     REAL REF_PAR(7)                !General setting parameters after refine
C     REAL RFCL_PAR(6)               !Cell parameters after refine
C     REAL DSTOR_PAR(2*MAX_SPD+4)    !Distortion parameters after refine
C     REAL VALUE                     !Parameter value
C     REAL CELL(6)                   !Real cell parameters
C     REAL RCELL(6)                  !Reciprocal cell params (expanded form)
C     REAL CELL_PARAMS(6)            !Recip. cell params. (contracted form)
C     REAL RTOD                      !Conversion factor, radians to degrees
C     DOUBLE PRECISION X_VEC(MAX_N)  !Shifts vector
C     DOUBLE PRECISION E_VEC(MAX_N)  !Errors vector
C     DOUBLE PRECISION FMIN          !Minimum value sum of squares
C     DOUBLE PRECISION FVECC(MAX_M)  !Vector of calculated values
C     DOUBLE PRECISION
C     DOUBLE PRECISION LSR_SHFLM     !Shift limit for LSQ refinement
C     DOUBLE PRECISION LSR_DSTEP     !Step for LSQ refinement
C     DOUBLE PRECISION LSR_DAMP      !Damping factor for LSQ refinement
C     CHARACTER*2 VALSTR             !Parameter value string (dummy)
C     CHARACTER*2 ERRSTR             !Error string (dummy)
C
      DATA NCP/6,4,3,2,2,2,1/
      DATA RTOD/57.29578/
C     
C-------------------------------------------------------------------------------
C
C====== Initialisations
C
      IERR = 0
      KRPACK = IPACK
      KRPLATE = IPLATE
      M = NUM_REF_OBS
      IF (IRFTYP.EQ.2) THEN
         POWELL_ACC = TUNE(1)
      ELSE
         LSR_SHFLM = TUNE(1)
         LSR_DSTEP = TUNE(2)
         LSR_DAMP = TUNE(3)
      END IF
      CALL LDM_GET_SYST (VALSTR, ISYST, LFLAG)
      CALL LDM_GET_DSTOR (VALSTR, DISTOR_TYPE, LFLAG)
      IF (DISTOR_TYPE.EQ.3) THEN
         CALL LDM_SPDXY (IPACK, IPLATE, SPD_XY,
     +                   NSPDX, SPDX_MIN, SPDX_MAX, SPDX,
     +                   NSPDY, SPDY_MIN, SPDY_MAX, SPDY,
     +                   MAX_SPD, IERR, ERRSTR)
         IF (NSPDX.GT.MAX_SPD.OR.NSPDY.GT.MAX_SPD) THEN
            ERRSTR = 'Polynomial order too high for refinement routine'
            IERR = -1
            RETURN
         END IF
      END IF
      DO 10 I=1,MAX_N
         X_VEC(I) = 0.0
         E_VEC(I) = 0.0
         SCALING(I) = 1.0
10     CONTINUE
C
C====== Set general setting parameters
C
      CALL LDM_GET_PHIX (IPACK, VALUE, LFLAG, ND, LERR, ERRSTR)
      X_START(1) = VALUE
      CALL LDM_GET_PHIY (IPACK, VALUE, LFLAG, ND, LERR, ERRSTR)
      X_START(2) = VALUE
      CALL LDM_GET_PHIZ (IPACK, VALUE, LFLAG, ND, LERR, ERRSTR)
      X_START(3) = VALUE
      CALL LDM_GET_CTOF (IPACK, IPLATE, VALUE, LFLAG, ND, 
     +                   LERR, ERRSTR)
      X_START(4) = VALUE
      CALL LDM_GET_XC (IPACK, IPLATE, VALUE, LFLAG, ND, 
     +                 LERR, ERRSTR)
      X_START(5) = VALUE
      CALL LDM_GET_YC (IPACK, IPLATE, VALUE, LFLAG, ND, 
     +                 LERR, ERRSTR)
      X_START(6) = VALUE
      CALL LDM_GET_WC (IPACK, IPLATE, VALUE, LFLAG, ND, 
     +                 LERR, ERRSTR)
      X_START(7) = VALUE
C
C====== Set cell parameters
C
      CALL LDM_GET_A (CELL(1), LFLAG, ND)
      CALL LDM_GET_B (CELL(2), LFLAG, ND)
      CALL LDM_GET_C (CELL(3), LFLAG, ND)
      CALL LDM_GET_ALPHA (CELL(4), LFLAG, ND)
      CALL LDM_GET_BETA (CELL(5), LFLAG, ND)
      CALL LDM_GET_GAMMA (CELL(6), LFLAG, ND)
      CALL RECCEL (RCELL,CELL,1.0)
      CALL CELLCX (ISYST,CELL_PARAMS,RCELL,2,2)
      DO 20 I=1,NCP(ISYST)
         XCELL_START(I) = CELL_PARAMS(I)
20    CONTINUE
C
C====== Set distortion parameters
C
      CALL LDM_GET_YSCAL (IPACK, IPLATE, VALUE, LFLAG, ND, 
     +                    LERR, ERRSTR)
      XDSTOR_START(1) = VALUE
      CALL LDM_GET_TWIST (IPACK, IPLATE, VALUE, LFLAG, ND, 
     +                   LERR, ERRSTR)
      XDSTOR_START(2) = VALUE
      CALL LDM_GET_TILT (IPACK, IPLATE, VALUE, LFLAG, ND, 
     +                   LERR, ERRSTR)
      XDSTOR_START(3) = VALUE
      IF (DISTOR_TYPE.EQ.1) THEN
         CALL LDM_GET_BULGE (IPACK, IPLATE, VALUE, LFLAG, ND, 
     +                       LERR, ERRSTR)
         XDSTOR_START(4) = VALUE
      ELSE IF (DISTOR_TYPE.EQ.2) THEN
         CALL LDM_GET_ROFF (IPACK, IPLATE, VALUE, LFLAG, ND, 
     +                      LERR, ERRSTR)
         XDSTOR_START(4) = VALUE
         CALL LDM_GET_TOFF (IPACK, IPLATE, VALUE, LFLAG, ND, 
     +                      LERR, ERRSTR)
         XDSTOR_START(5) = VALUE
      ELSE IF (DISTOR_TYPE.EQ.3) THEN
         XDSTOR_START(4) = SPD_XY
         DO 30 I = 1, NSPDX
            XDSTOR_START(I+4) = SPDX(I)
30       CONTINUE
         DO 40 I = 1, NSPDY
            XDSTOR_START(NSPDX+I+4) = SPDY(I)
40       CONTINUE
      END IF
      LOC_XCENF = X_CEN_F
      LOC_YCENF = Y_CEN_F
C
C====== Set up internal refinement flags (a flag holds the position in the
C       list of the parameters to be refined if the parameter is to be
C       refined or 0 if it is not to be refined
C
      N = 0
      DO 100 I=1,7
         REF_FLAG(I) = 0
         IF (IFLAG(I).EQ.1) THEN
            N = N + 1
            REF_FLAG(I) = N
         END IF
100   CONTINUE 
      REFINE_CELL = .FALSE.
      DO 110 I = 1, NCP(ISYST)
         RFCL_FLAG(I) = 0
         IF (JFLAG(I).EQ.1) THEN
            N = N + 1
            RFCL_FLAG(I) = N
            SCALING(N) = 1.0E-03
            REFINE_CELL = .TRUE.
         END IF
110   CONTINUE
      REFINE_DSTOR = .FALSE.
      REFINE_SPD = .FALSE.
      DO 115 I=1,(2*MAX_SPD+4)
      DSTOR_FLAG(I)=0
  115 CONTINUE
      DO 120 I = 1, 3
C        IF ((I.GT.1).OR.(DISTOR_TYPE.NE.4)) THEN
           IF (KFLAG(I).EQ.1) THEN
              N = N + 1
              DSTOR_FLAG(I) = N
              REFINE_DSTOR = .TRUE.
           END IF
C        END IF
120   CONTINUE
      IF (DISTOR_TYPE.EQ.1) THEN
         IF (KFLAG(4).EQ.1) THEN
            N = N + 1
            DSTOR_FLAG(4) = N
            REFINE_DSTOR = .TRUE.
         END IF
      ELSE IF (DISTOR_TYPE.EQ.2) THEN
         DO 130 I = 4, 5
            IF (KFLAG(I).EQ.1) THEN
               N = N + 1
               DSTOR_FLAG(I) = N
               REFINE_DSTOR = .TRUE.
            END IF
130      CONTINUE
      ELSE IF (DISTOR_TYPE.EQ.3) THEN
         IF (KFLAG(4).EQ.1) THEN
            N = N + 1
            DSTOR_FLAG(4) = N
            REFINE_SPD = .TRUE.
         END IF
         DO 140 I = 1, NSPDX
            IF (KFLAG(5).EQ.1) THEN
               N = N + 1
               DSTOR_FLAG(I+4) = N
               REFINE_SPD = .TRUE.
               SCALING(N) = 10.0
            END IF
140      CONTINUE
         DO 150 I = 1, NSPDY
            IF (KFLAG(6).EQ.1) THEN
               N = N + 1
               DSTOR_FLAG(I+NSPDX+4) = N
               REFINE_SPD = .TRUE.
               SCALING(N) = 10.0
            END IF
150      CONTINUE
      END IF

      FIRST = .TRUE.
      CHEB_OK = .TRUE.
      IF (IRFTYP.EQ.1) THEN
         CALL LSQMIN (M,N,X_VEC,E_VEC,LSR_SHFLM,50,LSR_DSTEP,LSR_DAMP,
     +                LFN_CALFUN,FMIN,FVECC,IFL,W)
         IF (DISTOR_TYPE.EQ.3.AND..NOT.CHEB_OK) THEN
            IERR = 3
            MESSAGE = '**Spots found outside Chebyshev limits**'
         ELSE IF (IFL.EQ.-1) THEN
            IERR = 1
            MESSAGE = '**Refinement not complete after 50 iterations**'
         ELSE IF (IFL.EQ.-2) THEN
            IERR = 2
            MESSAGE = '**Normal matrix is singular**'
         ELSE
            MESSAGE = 'Refinement completed after'
            CALL STR_ADDI(MESSAGE, IFL,1)
            CALL STR_ADDS(MESSAGE,'iterations',1)
         END IF
      ELSE IF (IRFTYP.EQ.2) THEN
         CALL SSQMIN (M,N,X_VEC,E_VEC,POWELL_ACC,50,LFN_CALFUN,FMIN,
     +                FVECC,IFL,W)
         IF (DISTOR_TYPE.EQ.3.AND..NOT.CHEB_OK) THEN
            IERR = 3
            MESSAGE = '**Spots found outside Chebyshev limits**'
         ELSE IF (IFL.EQ.-1) THEN
            IERR = 2
            MESSAGE = '**One or more failures in line minimisation**'
         ELSE IF (IFL.EQ.-2) THEN
            MESSAGE = '**Refinement not complete after 50 iterations**'
            IERR = 1
         ELSE
            MESSAGE = 'Refinement completed after'
            CALL STR_ADDI(MESSAGE, IFL,1)
            CALL STR_ADDS(MESSAGE,'iterations',1)
         END IF
      END IF
      RMS = FMIN/M
      RMS = SQRT(RMS)
C
C====== Get values of updated parameters
C
      SOVEMX = 0.0
      SOVEAV = 0.0
      SOVSUM = 0.0
      SOVN = 0.0
      DO 200 I=1,7
         IF (REF_FLAG(I).GT.0) THEN
            REF_PAR(I) = X_VEC(REF_FLAG(I))*SCALING(REF_FLAG(I))
     +                    + X_START(I)
            IF (E_VEC(REF_FLAG(I)).GT.0.0D0) THEN
              R1 = X_VEC(REF_FLAG(I))/E_VEC(REF_FLAG(I))
              IF (R1.LT.0.0) R1 = -R1
              IF (R1.GT.SOVEMX) SOVEMX = R1
              SOVSUM=SOVSUM + R1
              SOVN = SOVN + 1.0
            ENDIF
         ELSE
            REF_PAR(I) = X_START(I)
         END IF
200   CONTINUE
      DO 210 I=1,NCP(ISYST)
         IF (RFCL_FLAG(I).GT.0) THEN
            RFCL_PAR(I) = X_VEC(RFCL_FLAG(I))*SCALING(RFCL_FLAG(I))
     +                    + XCELL_START(I)
            IF (E_VEC(RFCL_FLAG(I)).GT.0.0D0) THEN
              R1 = X_VEC(RFCL_FLAG(I))/E_VEC(RFCL_FLAG(I))
              IF (R1.LT.0.0) R1 = -R1
              IF (R1.GT.SOVEMX) SOVEMX = R1
              SOVSUM=SOVSUM + R1
              SOVN = SOVN + 1.0
            ENDIF
         ELSE
            RFCL_PAR(I) = XCELL_START(I)
         END IF
210   CONTINUE
      IF (DISTOR_TYPE.EQ.1) THEN
         ND = 4
      ELSE IF (DISTOR_TYPE.EQ.2) THEN
         ND = 5
      ELSE IF (DISTOR_TYPE.EQ.3) THEN
         ND = 4 + NSPDX + NSPDY
      ELSE IF (DISTOR_TYPE.EQ.4) THEN
         ND = 3
      END IF
      DO 220 I=1, ND
         IF (DSTOR_FLAG(I).GT.0) THEN
            DSTOR_PAR(I) = X_VEC(DSTOR_FLAG(I))*SCALING(DSTOR_FLAG(I))
     +                     + XDSTOR_START(I)
            IF (E_VEC(DSTOR_FLAG(I)).GT.0.0D0) THEN
              R1 = X_VEC(DSTOR_FLAG(I))/E_VEC(DSTOR_FLAG(I))
              IF (R1.LT.0.0) R1 = -R1
              IF (R1.GT.SOVEMX) SOVEMX = R1
              SOVSUM=SOVSUM + R1
              SOVN = SOVN + 1.0
            ENDIF
            ELSE
               DSTOR_PAR(I) = XDSTOR_START(I)
            END IF
220     CONTINUE
      IF (SOVN.GT.0.0) SOVEAV = SOVSUM/SOVN
C
C====== Update LDM
C
      IF (IFLAG(1).EQ.1) CALL LDM_SET_PHIX (IPACK, REF_PAR(1), 
     +    KERR, ERRSTR)
      IF (IFLAG(2).EQ.1) CALL LDM_SET_PHIY (IPACK, REF_PAR(2), 
     +    KERR, ERRSTR)
      IF (IFLAG(3).EQ.1) CALL LDM_SET_PHIZ (IPACK, REF_PAR(3), 
     +    KERR, ERRSTR)
      IF (IFLAG(4).EQ.1) CALL LDM_SET_CTOF (IPACK, IPLATE, REF_PAR(4), 
     +    KERR, ERRSTR)
      IF (IFLAG(5).EQ.1) CALL LDM_SET_XC (IPACK, IPLATE, REF_PAR(5), 
     +    KERR, ERRSTR)
      IF (IFLAG(6).EQ.1) CALL LDM_SET_YC (IPACK, IPLATE, REF_PAR(6), 
     +    KERR, ERRSTR)
      IF (IFLAG(7).EQ.1) CALL LDM_SET_WC (IPACK, IPLATE, REF_PAR(7), 
     +    KERR, ERRSTR)

      CALL CELLCX (ISYST, RFCL_PAR, RCELL, 1, 2)
      CALL RECCEL (CELL, RCELL, 1.0)
      IF (ISYST.EQ.1) THEN
         IF (JFLAG(1).EQ.1) CALL LDM_SET_A (CELL(1), KERR, ERRSTR)
         IF (JFLAG(2).EQ.1) CALL LDM_SET_B (CELL(2), KERR, ERRSTR)
         IF (JFLAG(3).EQ.1) CALL LDM_SET_C (CELL(3), KERR, ERRSTR)
         IF (JFLAG(4).EQ.1) CALL LDM_SET_ALPHA (CELL(4), KERR, ERRSTR)
         IF (JFLAG(5).EQ.1) CALL LDM_SET_BETA (CELL(5), KERR, ERRSTR)
         IF (JFLAG(6).EQ.1) CALL LDM_SET_GAMMA (CELL(6), KERR, ERRSTR)
      ELSE IF (ISYST.EQ.2) THEN
         IF (JFLAG(1).EQ.1) CALL LDM_SET_A (CELL(1), KERR, ERRSTR)
         IF (JFLAG(2).EQ.1) CALL LDM_SET_B (CELL(2), KERR, ERRSTR)
         IF (JFLAG(3).EQ.1) CALL LDM_SET_C (CELL(3), KERR, ERRSTR)
         IF (JFLAG(4).EQ.1) CALL LDM_SET_BETA (CELL(5), KERR, ERRSTR)
      ELSE IF (ISYST.EQ.3) THEN
         IF (JFLAG(1).EQ.1) CALL LDM_SET_A (CELL(1), KERR, ERRSTR)
         IF (JFLAG(2).EQ.1) CALL LDM_SET_B (CELL(2), KERR, ERRSTR)
         IF (JFLAG(3).EQ.1) CALL LDM_SET_C (CELL(3), KERR, ERRSTR)
      ELSE IF (ISYST.EQ.4) THEN
         IF (JFLAG(1).EQ.1) CALL LDM_SET_A (CELL(1), KERR, ERRSTR)
         IF (JFLAG(1).EQ.1) CALL LDM_SET_B (CELL(2), KERR, ERRSTR)
         IF (JFLAG(2).EQ.1) CALL LDM_SET_C (CELL(3), KERR, ERRSTR)
      ELSE IF (ISYST.EQ.5) THEN
         IF (JFLAG(1).EQ.1) CALL LDM_SET_A (CELL(1), KERR, ERRSTR)
         IF (JFLAG(1).EQ.1) CALL LDM_SET_B (CELL(2), KERR, ERRSTR)
         IF (JFLAG(2).EQ.1) CALL LDM_SET_C (CELL(3), KERR, ERRSTR)
      ELSE IF (ISYST.EQ.6) THEN
         IF (JFLAG(1).EQ.1) CALL LDM_SET_A (CELL(1), KERR, ERRSTR)
         IF (JFLAG(1).EQ.1) CALL LDM_SET_B (CELL(2), KERR, ERRSTR)
         IF (JFLAG(1).EQ.1) CALL LDM_SET_C (CELL(3), KERR, ERRSTR)
         IF (JFLAG(2).EQ.1) CALL LDM_SET_ALPHA (CELL(4), KERR, ERRSTR)
         IF (JFLAG(2).EQ.1) CALL LDM_SET_BETA (CELL(5), KERR, ERRSTR)
         IF (JFLAG(2).EQ.1) CALL LDM_SET_GAMMA (CELL(6), KERR, ERRSTR)
      ELSE IF (ISYST.EQ.7) THEN
         IF (JFLAG(1).EQ.1) CALL LDM_SET_A (CELL(1), KERR, ERRSTR)
         IF (JFLAG(1).EQ.1) CALL LDM_SET_B (CELL(2), KERR, ERRSTR)
         IF (JFLAG(1).EQ.1) CALL LDM_SET_C (CELL(3), KERR, ERRSTR)
      END IF

      IF (DISTOR_TYPE.EQ.1) THEN
         IF (KFLAG(1).EQ.1) CALL LDM_SET_YSCAL (IPACK, IPLATE, 
     +       DSTOR_PAR(1), KERR, ERRSTR)
         IF (KFLAG(2).EQ.1) CALL LDM_SET_TWIST (IPACK, IPLATE, 
     +       DSTOR_PAR(2), KERR, ERRSTR)        
         IF (KFLAG(3).EQ.1) CALL LDM_SET_TILT (IPACK, IPLATE, 
     +       DSTOR_PAR(3), KERR, ERRSTR)
         IF (KFLAG(4).EQ.1) CALL LDM_SET_BULGE (IPACK, IPLATE, 
     +       DSTOR_PAR(4), KERR, ERRSTR)
      ELSE IF (DISTOR_TYPE.EQ.2) THEN
         IF (KFLAG(1).EQ.1) CALL LDM_SET_YSCAL (IPACK, IPLATE, 
     +       DSTOR_PAR(1), KERR, ERRSTR)
         IF (KFLAG(2).EQ.1) CALL LDM_SET_TWIST (IPACK, IPLATE, 
     +       DSTOR_PAR(2), KERR, ERRSTR)        
         IF (KFLAG(3).EQ.1) CALL LDM_SET_TILT (IPACK, IPLATE, 
     +       DSTOR_PAR(3), KERR, ERRSTR)
         IF (KFLAG(4).EQ.1) CALL LDM_SET_ROFF (IPACK, IPLATE, 
     +       DSTOR_PAR(4), KERR, ERRSTR)
         IF (KFLAG(5).EQ.1) CALL LDM_SET_TOFF (IPACK, IPLATE, 
     +       DSTOR_PAR(5), KERR, ERRSTR)
      ELSE IF (DISTOR_TYPE.EQ.3) THEN
         IF (KFLAG(1).EQ.1) CALL LDM_SET_YSCAL (IPACK, IPLATE, 
     +       DSTOR_PAR(1), KERR, ERRSTR)
         IF (KFLAG(2).EQ.1) CALL LDM_SET_TWIST (IPACK, IPLATE, 
     +       DSTOR_PAR(2), KERR, ERRSTR)        
         IF (KFLAG(3).EQ.1) CALL LDM_SET_TILT (IPACK, IPLATE, 
     +       DSTOR_PAR(3), KERR, ERRSTR)
         IF (KFLAG(4).EQ.1) CALL LDM_SET_SPDXY (IPACK, IPLATE, 
     +       DSTOR_PAR(4), KERR, ERRSTR)
         IF (KFLAG(5).EQ.1) THEN
            DO 230 I = 1, NSPDX
               CALL LDM_SET_SPDX (I, IPACK, IPLATE, DSTOR_PAR(I+4),
     +                      KERR, ERRSTR)
230         CONTINUE
         END IF
         IF (KFLAG(6).EQ.1) THEN
            DO 240 I = 1, NSPDY
               CALL LDM_SET_SPDY (I, IPACK, IPLATE,
     +                            DSTOR_PAR(I+NSPDX+4),       
     +                            KERR, ERRSTR)
240         CONTINUE
         END IF
      ELSE IF (DISTOR_TYPE.EQ.4) THEN
         IF (KFLAG(1).EQ.1) CALL LDM_SET_YSCAL (IPACK, IPLATE, 
     +       DSTOR_PAR(1), KERR, ERRSTR)
         IF (KFLAG(2).EQ.1) CALL LDM_SET_TWIST (IPACK, IPLATE, 
     +       DSTOR_PAR(2), KERR, ERRSTR)        
         IF (KFLAG(3).EQ.1) CALL LDM_SET_TILT (IPACK, IPLATE, 
     +       DSTOR_PAR(3), KERR, ERRSTR)
      END IF
      RETURN
      END
